@@ -26,7 +26,8 @@ import io
 from ..parser import Tree
 from .colors import color
 from .defs import gradient_or_pattern, parse_def
-from .helpers import node_format, transform
+from .helpers import (
+    node_format, transform, normalize, filter_fill_or_stroke, PointError)
 from .path import PATH_TAGS
 from .tags import TAGS
 from .units import size
@@ -90,7 +91,6 @@ class Surface(object):
 
         """
         self.cairo = None
-        self.context = None
         self.context_width, self.context_height = None, None
         self.font_size = None
         self.cursor_position = 0, 0
@@ -221,11 +221,22 @@ class Surface(object):
         if join_cap == "bevel":
             self.context.set_line_join(cairo.LINE_JOIN_BEVEL)
 
+        dash_array = normalize(node.get("stroke-dasharray", "")).split()
+        if dash_array:
+            dashes = [size(self, dash) for dash in dash_array]
+            if sum(dashes):
+                offset = size(self, node.get("stroke-dashoffset"))
+                self.context.set_dash(dashes, offset)
+
         miter_limit = float(node.get("stroke-miterlimit", 4))
         self.context.set_miter_limit(miter_limit)
 
         if node.tag in TAGS:
-            TAGS[node.tag](self, node)
+            try:
+                TAGS[node.tag](self, node)
+            except PointError:
+                # Error in point parsing, do nothing
+                pass
 
         # Get stroke and fill opacity
         stroke_opacity = float(node.get("stroke-opacity", 1))
@@ -237,23 +248,31 @@ class Surface(object):
 
         if stroke_and_fill and visible:
             # Fill
-            if "url(#" in node.get("fill", ""):
-                gradient_or_pattern(self, node)
+            if "url(#" in (node.get("fill") or ""):
+                name = filter_fill_or_stroke(node.get("fill"))
+                gradient_or_pattern(self, node, name)
             else:
                 if node.get("fill-rule") == "evenodd":
                     self.context.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)
                 self.context.set_source_rgba(
                     *color(node.get("fill", "black"), fill_opacity))
-                self.context.fill_preserve()
+            self.context.fill_preserve()
 
             # Stroke
             self.context.set_line_width(size(self, node.get("stroke-width")))
-            self.context.set_source_rgba(
-                *color(node.get("stroke"), stroke_opacity))
+            if "url(#" in (node.get("stroke") or ""):
+                name = filter_fill_or_stroke(node.get("stroke"))
+                gradient_or_pattern(self, node, name)
+            else:
+                self.context.set_source_rgba(
+                    *color(node.get("stroke"), stroke_opacity))
             self.context.stroke()
+        elif not visible:
+            self.context.new_path()
 
         # Draw children
-        if display:
+        if display and node.tag not in (
+                "linearGradient", "radialGradient", "marker", "pattern"):
             for child in node.children:
                 self.draw(child, stroke_and_fill)
 
